@@ -26,6 +26,7 @@
 
 #include <kdecorationbridge.h>
 #include <KDE/KActionCollection>
+#include <kdeversion.h>
 
 #include <qpixmap.h>
 #include <qwidget.h>
@@ -55,6 +56,7 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 
 	enum Type
 	{
+	    Normal2D,
 	    Normal,
 	    Default,
 	    DefaultActive
@@ -62,7 +64,7 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 
     public:
 	Window (WId parentId, WId clientId, WId frame, Type type,
-		int x = 0, int y = 0, int w = 1, int h = 1);
+		QRect geometry = QRect ());
 	~Window (void);
 
 	virtual bool isActive (void) const;
@@ -83,10 +85,13 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 	virtual NET::WindowType
 	    windowType (unsigned long supported_types) const;
 	virtual QIcon icon (void) const;
+	virtual QIcon icon (int idx) const;
 	virtual QString caption (void) const;
+	virtual QString caption (int idx) const;
 	virtual void processMousePressEvent (QMouseEvent *);
 	virtual void showWindowMenu (const QRect &);
 	virtual void showWindowMenu (const QPoint &);
+	virtual void showWindowMenu (const QPoint &, long int id);
 	virtual void performWindowOperation (WindowOperation);
 	virtual void setMask (const QRegion &, int);
 	virtual bool isPreview (void) const;
@@ -111,8 +116,35 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 
 	/* unstable API */
 	virtual bool compositingActive () const;
-	
-	virtual bool eventFilter (QObject* o, QEvent* e);
+
+#if KDE_IS_VERSION(4,3,90)
+	virtual QRect transparentRect () const;
+
+	virtual bool isClientGroupActive ();
+	virtual QList<ClientGroupItem> clientGroupItems () const;
+	virtual long itemId (int index);
+	virtual int visibleClientGroupItem ();
+	virtual void setVisibleClientGroupItem (int index);
+	virtual void moveItemInClientGroup (int index, int before);
+	virtual void moveItemToClientGroup (long itemId, int before);
+	virtual void removeFromClientGroup (int index, const QRect& newGeom);
+	virtual void closeClientGroupItem (int index);
+	virtual void closeAllInClientGroup ();
+	virtual void displayClientMenu (int index, const QPoint& pos);
+
+	virtual WindowOperation
+	    buttonToWindowOperation(Qt::MouseButtons button);
+#endif
+
+	long tabId (int idx) const;
+	long currentTabId () const;
+	void setCurrentTab (long id);
+	void tab_A_before_B (long A, long B);
+	void tab_A_behind_B (long A, long B);
+	void untab (long id, const QRect& newGeom);
+	void closeTab (long id);
+	void closeTabGroup ();
+	int tabCount () const;
 
 	void handleActiveChange (void);
 	void updateFrame (WId frame);
@@ -159,9 +191,6 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 	    return mPixmap;
 	}
 	
-	bool handleMap (void);
-	bool handleConfigure (QSize size);
-	
 	decor_extents_t *border (void)
 	{
 	    return &mBorder;
@@ -181,6 +210,7 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 	    return mFakeRelease;
 	}
 	
+	virtual bool eventFilter (QObject *o, QEvent *e);
 
     private:
 	void createDecoration (void);
@@ -189,14 +219,17 @@ class Window: public QObject, public KDecorationBridgeUnstable {
 				 int bottomOffset,
 				 int leftOffset,
 				 int rightOffset);
-	void updateProperty (void);
+	long * updateProperty (long *, unsigned int);
 	void getWindowProtocols (void);
+
+	Options::MouseCommand buttonToCommand (Qt::MouseButtons button);
 	void performMouseCommand (KWD::Options::MouseCommand command,
-				  QMouseEvent		     *qme);
+				  QMouseEvent                *qme);
 	NET::Direction positionToDirection (int pos);
 	Cursor positionToCursor (QPoint pos);
 
     private slots:
+	void resizeDecorationTimeout ();
 	void handlePopupActivated (QAction *action);
 	void handleOpacityPopupActivated (QAction *action);
 	void handleDesktopPopupActivated (QAction *action);
